from typing import List, Tuple

from base64 import b64decode

from gidgetlab.routing import Router
from gidgetlab.sansio import Event
from gidgetlab.abc import GitLabAPI

from ..helpers.git import get_remote_git_head_hash, get_submodules

router_submodule = Router()


def log(event: Event, string: str) -> None:
    print("[submodule, {}, {}] {}".format(event.project_id, event.object_attributes['iid'], string))


def get_project_base_url(event: Event) -> str:
    """Get the base url for the project of the given event"""
    return f"/projects/{event.project_id}"


def get_mr_base_url(event: Event) -> str:
    """Get the base url for the mr of the given event"""
    return (get_project_base_url(event) +
            f"/merge_requests/{event.object_attributes['iid']}")


async def get_changed_submodules(gl: GitLabAPI, mr_base_url: str) -> List[Tuple[str, str]]:
    """Get a list of changed submodules for the mr"""
    changes_dict = await gl.getitem(mr_base_url + "/changes")
    return [(change["new_path"], change["diff"].split("Subproject commit")[-1].strip()) for
            change in
            changes_dict["changes"] if
            "Subproject commit" in change["diff"]]


async def get_gitmodules(gl: GitLabAPI, event: Event, commit: str) -> str:
    """Get the content of the .gitmodules file"""
    url = (get_project_base_url(event) +
           f"/repository/files/.gitmodules?ref={commit}")
    file_dict = await gl.getitem(url)
    if file_dict["encoding"] != "base64":
        raise Exception("Only support base64 encoded text")
    return b64decode(file_dict["content"]).decode("utf-8")


@router_submodule.register("Merge Request Hook", action="open")
@router_submodule.register("Merge Request Hook", action="update")
async def check_submodules_are_up_to_date(event: Event, gl: GitLabAPI):
    base_url = get_mr_base_url(event)
    commit_to_look_at = event.object_attributes['last_commit']['id']
    changed_submodules = await get_changed_submodules(gl, base_url)

    log(event, "Found changed submodules: {}".format(repr(changed_submodules)))

    if len(changed_submodules) == 0:
        return None

    gitmodules_content = await get_gitmodules(gl, event, commit_to_look_at)
    submodules = get_submodules(gitmodules_content)
    submodules_by_path = dict([(sm.path, sm) for sm in submodules])

    result_string = "The submodule analysis has found: \n"

    for path, hash in changed_submodules:
        if path not in submodules_by_path:
            raise Exception("Submodule not in list of all submodules")
        sm = submodules_by_path[path]
        newest_hash = get_remote_git_head_hash(sm.url)

        log(event, "SM: {}, to:{}, newest:{}".format(sm.name, hash, newest_hash))

        if newest_hash == hash:
            msg = "- Updates submodule {} to upstream HEAD. \n"
        else:
            msg = "- WARNING: The submodule {} is changed but not to upstream HEAD. \n"

        result_string += msg.format(sm.name)

    log(event, "Posting response")

    url = base_url + "/notes"
    await gl.post(url, data={"body": result_string})
