from typing import KeysView, Optional, List, Set
from dataclasses import dataclass

import re

import subprocess

FATAL_REPO_NOT_FOUND = "fatal: Could not read from remote repository"


def get_remote_git_head_hash(git_url: str) -> str:
    """Get the hash of the neweset commit of the primary branch of the repo"""
    command = ["git", "ls-remote", git_url, "HEAD"]
    try:
        output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        return output.decode('utf-8').split()[0]
    except subprocess.CalledProcessError as e:
        if FATAL_REPO_NOT_FOUND in e.output.decode('utf-8') and "git@" in git_url:
            # Try again with a guessed https address
            new_url = re.sub(r"^git@(ssh.)?([0-9a-zA-Z._-]*):", r"https://\2/", git_url)
            new_url = git_url.replace("git@github.com:", "https://github.com/", 1)
            return get_remote_git_head_hash(new_url)
        raise e


@dataclass
class GitConfigEntry():
    """One entry in a git config style file"""
    section: str
    sub_section: Optional[str]
    name: str
    value: str


def _entry_from_line(line: str) -> GitConfigEntry:
    line_parts = line.split("=", maxsplit=1)
    identifier = line_parts[0]
    value = line_parts[1]
    identifier_parts = identifier.split(".", maxsplit=2)
    if len(identifier_parts) == 2:
        return GitConfigEntry(identifier_parts[0], None, identifier_parts[1], value)
    elif len(identifier_parts) == 3:
        return GitConfigEntry(identifier_parts[0], identifier_parts[1], identifier_parts[2], value)
    else:
        assert False


def list_git_config_entries(git_config: str) -> List[GitConfigEntry]:
    """Get all config entries in the given git config string"""
    command = ["git", "config", "-f", "-", "-l"]
    output = subprocess.check_output(command, input=git_config.encode("utf-8"))
    result = list()
    for line in output.decode("utf-8").split("\n"):
        if len(line) != 0:
            result.append(_entry_from_line(line))
    return result


@dataclass
class GitSubmodule():
    """One submodule"""
    name: str
    path: str
    url: str


def _set_of_s(keys: KeysView) -> Set[str]:
    return set([key for key in keys if key is not None])


def get_submodules(gitsubmodules_file_content: str) -> List[GitSubmodule]:
    """Get all submodules from the given content of the gitmodules file"""
    entries = list_git_config_entries(gitsubmodules_file_content)
    paths = dict()
    urls = dict()
    for entry in entries:
        if entry.section != "submodule":
            raise Exception("Only expected submodule configs")
        if entry.name == "path":
            if entry.sub_section in paths:
                raise Exception("Duplicate path for: {}".format(entry.sub_section))
            paths[entry.sub_section] = entry.value
        elif entry.name == "url":
            if entry.sub_section in urls:
                raise Exception("Duplicate url for: {}".format(entry.sub_section))
            urls[entry.sub_section] = entry.value
        else:
            raise Exception("Only expected the keys path and url")

    result = list()
    all_keys: Set[str] = _set_of_s(paths.keys()).union(_set_of_s(paths.keys()))
    for name in all_keys:
        if name not in paths:
            raise Exception("Missing path for: {}".format(name))
        if name not in urls:
            raise Exception("Missing url for: {}".format(name))
        result.append(GitSubmodule(name, paths[name], urls[name]))
    return result
